import os
import torch
from pytorchDL.layers import Conv2dLayer, Conv2dTransposeLayer


def get_script_path():
    return os.path.realpath(__file__)


class Conv2dBlock(torch.nn.Module):

    def __init__(self, in_features, out_features, kernel_size=(3, 3), stride=(1, 1)):
        super(Conv2dBlock, self).__init__()

        self.out_features = out_features
        self._cn1 = Conv2dLayer(in_features, out_features, activation='relu',
                                kernel_size=kernel_size, stride=stride, padding=1)

        self._cn2 = Conv2dLayer(out_features, out_features, activation='relu',
                                kernel_size=kernel_size, stride=stride, padding=1)

    def forward(self, input_tensor):
        x = self._cn1(input_tensor)
        x = self._cn2(x)
        return x


class UNet(torch.nn.Module):

    def __init__(self, input_channels, output_channels, **kwargs):
        super(UNet, self).__init__()

        self.cn1 = Conv2dBlock(input_channels, 32)
        self.mp1 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.cn2 = Conv2dBlock(32, 64)
        self.mp2 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.cn3 = Conv2dBlock(64, 128)
        self.mp3 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.cn4 = Conv2dBlock(128, 256)
        self.mp4 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.drop1 = torch.nn.Dropout2d(p=0.3)
        self.cn5 = Conv2dBlock(256, 512)

        self.up_cn1 = Conv2dTransposeLayer(512, 256, activation='relu')
        self.cn6 = Conv2dLayer(512, 256, activation='relu')

        self.up_cn2 = Conv2dTransposeLayer(256, 128, activation='relu')
        self.cn7 = Conv2dLayer(256, 128, activation='relu')

        self.up_cn3 = Conv2dTransposeLayer(128, 64, activation='relu')
        self.cn8 = Conv2dLayer(128, 64, activation='relu')

        self.up_cn4 = Conv2dTransposeLayer(64, 32, activation='relu')
        self.cn9 = Conv2dLayer(64, 32, activation='relu')

        self.cn10 = Conv2dLayer(32, output_channels, activation=None, kernel_size=(3, 3), padding=1)

    def forward(self, input_tensor):

        x1 = self.cn1(input_tensor)
        dw_x1 = self.mp1(x1)

        x2 = self.cn2(dw_x1)
        dw_x2 = self.mp2(x2)

        x3 = self.cn3(dw_x2)
        dw_x3 = self.mp3(x3)

        x4 = self.cn4(dw_x3)
        dw_x4 = self.mp4(x4)

        dw_x4 = self.drop1(dw_x4)
        x5 = self.cn5(dw_x4)

        up_x5 = self.up_cn1(x5)
        x6 = torch.cat((x4, up_x5), dim=1)
        x6 = self.cn6(x6)

        up_x6 = self.up_cn2(x6)
        x7 = torch.cat((x3, up_x6), dim=1)
        x7 = self.cn7(x7)

        up_x7 = self.up_cn3(x7)
        x8 = torch.cat((x2, up_x7), dim=1)
        x8 = self.cn8(x8)

        up_x8 = self.up_cn4(x8)
        x9 = torch.cat((x1, up_x8), dim=1)
        x9 = self.cn9(x9)

        return self.cn10(x9)



import os
from shutil import rmtree

import unittest
import torch

import numpy as np
from pytorchDL.loggers import TensorboardLogger


class TestTensorboardLogger(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cwd = os.path.dirname(os.path.realpath(__file__))
        cls.test_log_dir = os.path.join(cwd, 'test_logs')
        cls.tb_logger = TensorboardLogger(log_dir=cls.test_log_dir)

    def test_log_scalar(self):

        log_data = [{'type': 'scalar', 'data': 1.2342, 'name': 'set_0'},
                    {'type': 'scalar', 'data': np.random.uniform(), 'name': 'set_1'}]

        self.tb_logger.log(log_data, stage='train', step=0)
        self.tb_logger.log(log_data, stage='train', step=1)

        log_data = [{'type': 'scalar', 'data': torch.Tensor([0.1312]), 'name': 'set_1'}]
        self.tb_logger.log(log_data, stage='val', step=0)

    def test_log_np_image(self):
        img = np.random.uniform(0, 1, size=(1, 3, 32, 32))  # rgb image
        log_data = [{'type': 'image', 'data': img, 'name': 'imgs_0'}]
        self.tb_logger.log(log_data, stage='train', step=0)

        img = np.random.uniform(0, 1, size=(5, 1, 32, 32))  #grayscale image
        log_data = [{'type': 'image', 'data': img, 'name': 'imgs_0'}]
        self.tb_logger.log(log_data, stage='train', step=1)

    def test_log_tensor_image(self):

        img = torch.from_numpy(np.random.uniform(0, 1, size=(5, 3, 32, 32)))
        log_data = [{'type': 'image', 'data': img, 'name': 'imgs_0'},
                    {'type': 'image', 'data': img, 'name': 'imgs_1'}]

        self.tb_logger.log(log_data, stage='train', step=0)

    def test_log_np_pointcloud(self):
        pcd_xyz = np.random.uniform(-5, 5, size=(5, 3, 30))
        pcd_rgb = np.random.uniform(0, 1, size=(5, 3, 30))
        pcd = np.concatenate((pcd_xyz, pcd_rgb), axis=1)
        log_data = [{'type': 'pointcloud', 'data': pcd, 'name': 'pcd_0'}]

        self.tb_logger.log(log_data, stage='test', step=10)

    def test_log_tensor_pointcloud(self):
        pcd_xyz = torch.from_numpy(np.random.uniform(-5, 5, size=(5, 3, 30)))
        pcd_rgb = torch.from_numpy(np.random.uniform(0, 1, size=(5, 3, 30)))
        pcd = torch.cat((pcd_xyz, pcd_rgb), dim=1)
        log_data = [{'type': 'pointcloud', 'data': pcd, 'name': 'pcd_0'}]

        self.tb_logger.log(log_data, stage='test', step=10)

    @classmethod
    def tearDownClass(cls) -> None:
        rmtree(cls.test_log_dir)



import torch


class ActivationLayer(torch.nn.Module):

    def __init__(self, activation, **kwargs):
        super(ActivationLayer, self).__init__()

        activation = activation.lower()
        if activation == 'relu':
            self._act = torch.nn.ReLU(**kwargs)
        elif activation == 'leaky_relu':
            self._act = torch.nn.LeakyReLU(**kwargs)
        else:
            raise NotImplementedError('Error: %s activation layer not implemented!' % activation.upper())

    def forward(self, input_tensor):
        return self._act(input_tensor)


class Conv2dLayer(torch.nn.Module):

    def __init__(self, in_features, out_features, activation, bn=False, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=True):
        super(Conv2dLayer, self).__init__()
        
        self.out_features = out_features
        layers = list()
        layers.append(torch.nn.Conv2d(in_features, out_features, kernel_size, stride, padding=padding, bias=bias))
        if bn:
            layers.append(torch.nn.BatchNorm2d(num_features=out_features))
        if activation is not None:
            layers.append(ActivationLayer(activation))

        self.layers = torch.nn.Sequential(*layers)

    def forward(self, input_tensor):
        return self.layers(input_tensor)


class Conv2dTransposeLayer(torch.nn.Module):

    def __init__(self, in_features, out_features, activation, bn=False, kernel_size=(3, 3), stride=(2, 2), padding=1, output_padding=1, bias=True):
        super(Conv2dTransposeLayer, self).__init__()

        self.out_features = out_features
        layers = list()
        layers.append(torch.nn.ConvTranspose2d(in_features, out_features, kernel_size, stride, padding=padding, output_padding=output_padding, bias=bias))
        if bn:
            layers.append(torch.nn.BatchNorm2d(num_features=out_features))
        if activation is not None:
            layers.append(ActivationLayer(activation))

        self.layers = torch.nn.Sequential(*layers)

    def forward(self, input_tensor):
        return self.layers(input_tensor)


class FullyConnectedLayer(torch.nn.Module):

    def __init__(self, in_features, out_features, activation, bn=False, bias=True):
        super(FullyConnectedLayer, self).__init__()

        self.out_features = out_features
        layers = list()
        layers.append(torch.nn.Linear(in_features, out_features, bias=bias))
        if bn:
            layers.append(torch.nn.BatchNorm1d(num_features=out_features))
        if activation is not None:
            layers.append(ActivationLayer(activation))

        self.layers = torch.nn.Sequential(*layers)

    def forward(self, input_tensor):
        return self.layers(input_tensor)


class ResidualConv2dLayer(torch.nn.Module):
    def __init__(self, in_features, activation, bn=False, kernel_size=(3, 3), stride=(1, 1), padding=1, bias=True):

        super(ResidualConv2dLayer, self).__init__()

        self.out_features = in_features
        self._cn0 = Conv2dLayer(in_features, in_features,
                                activation=activation, bn=bn, kernel_size=kernel_size,
                                stride=stride, padding=padding, bias=bias)

        self._cn1 = Conv2dLayer(in_features, in_features,
                                activation=None, bn=bn, kernel_size=kernel_size,
                                stride=1, padding=padding, bias=bias)

        self.act = ActivationLayer(activation) if activation is not None else None

    def forward(self, input_tensor):

        x = self._cn0(input_tensor)
        x = self._cn1(x)

        out = torch.add(input_tensor, x)
        if self.act is not None:
            out = self.act(out)
        return out



import numpy as np


def rgb_to_yuv(img_rgb):
    """ Converts a rgb image array of dims [H, W, CH] to YUV color space.
    Input RGB ranges are assumed to be in [0, 1]
    Output YUV values ranges are:
        Y: [0, 1]
        U: [-0.436, 0.436]
        V: [-0.615, 0.615]
    """

    r = img_rgb[..., 0]
    g = img_rgb[..., 1]
    b = img_rgb[..., 2]

    y = 0.299 * r + 0.587 * g + 0.114 * b
    u = 0.493 * (b - y)
    v = 0.877 * (r - y)

    img_yuv = np.stack((y, u, v), axis=2)
    return img_yuv


def yuv_to_rgb(img_yuv):
    """ Converts a YUV image array of dims [H, W, CH] to RGB color space.
    Input YUV ranges are assumed to be:
        Y: [0, 1]
        U: [-0.436, 0.436]
        V: [-0.615, 0.615]
    Output RGB range is [0, 1]
    """

    y = img_yuv[..., 0]
    u = img_yuv[..., 1]
    v = img_yuv[..., 2]

    r = y + 1.14 * v
    g = y - 0.396 * u - 0.581 * v
    b = y + 2.029 * u

    img_rgb = np.stack((r, g, b), axis=2)
    img_rgb = np.clip(img_rgb, 0, 1)
    return img_rgb




import torch
import os


def get_script_path():
    return os.path.realpath(__file__)


class Conv2dTransposeUnit(torch.nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=2, padding=1, output_padding=1, batch_norm=True, activation='relu', bias=False):
        super(Conv2dTransposeUnit, self).__init__()

        layers = [torch.nn.ConvTranspose2d(in_channels, out_channels, kernel_size,
                                           stride, padding=padding, output_padding=output_padding, bias=bias)]
        if batch_norm:
            layers.append(torch.nn.BatchNorm2d(num_features=out_channels, momentum=0.01, eps=5e-05))

        if activation == 'relu':
            layers.append(torch.nn.ReLU(inplace=True))

        self.layers = torch.nn.Sequential(*layers)

    def forward(self, input):
        return self.layers(input)


class Conv2dUnit(torch.nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, batch_norm=True, activation='leaky_relu', bias=False):
        super(Conv2dUnit, self).__init__()

        layers = [torch.nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding=1, bias=bias)]

        if batch_norm:
            layers.append(torch.nn.BatchNorm2d(num_features=out_channels, momentum=0.01, eps=5e-05))

        if activation == 'leaky_relu':
            layers.append(torch.nn.LeakyReLU(0.2, inplace=True))
        elif activation == 'relu':
            layers.append(torch.nn.ReLU(inplace=True))

        self.layers = torch.nn.Sequential(*layers)

    def forward(self, input):
        return self.layers(input)


class Generator(torch.nn.Module):

    def __init__(self, latent_vector_length, output_channels):
        super().__init__()
        self.dense = torch.nn.Linear(latent_vector_length, 8*8*1024)
        self.leaky_relu = torch.nn.LeakyReLU(0.2, inplace=True)

        self.ct1 = Conv2dTransposeUnit(1024, 512)
        self.ct2 = Conv2dTransposeUnit(512, 256)
        self.ct3 = Conv2dTransposeUnit(256, 128)
        self.ct4 = Conv2dTransposeUnit(128, 64)
        self.ct5 = Conv2dTransposeUnit(64, output_channels, stride=1, output_padding=0, batch_norm=False, activation=None)

    def forward(self, input_tensor):

        x = self.leaky_relu(self.dense(input_tensor))
        x = x.view(-1, 1024, 8, 8)
        x = self.ct1(x)
        x = self.ct2(x)
        x = self.ct3(x)
        x = self.ct4(x)
        x = self.ct5(x)
        x = torch.tanh(x)
        return x


class Discriminator(torch.nn.Module):

    def __init__(self, input_shape):
        super().__init__()

        self.cn1 = Conv2dUnit(input_shape[-1], 64, stride=2, activation='leaky_relu', batch_norm=False)
        self.cn2 = Conv2dUnit(64, 128, stride=2, activation='leaky_relu')
        self.cn3 = Conv2dUnit(128, 256, stride=2, activation='leaky_relu')
        self.cn4 = Conv2dUnit(256, 512, stride=2, activation='leaky_relu')
        self.cn5 = Conv2dUnit(512, 1024, stride=2, activation='leaky_relu')

        self.dense_size = int(1024 * input_shape[0] / 32 * input_shape[0] / 32)
        self.dense = torch.nn.Linear(self.dense_size, 1, bias=False)

    def forward(self, input_tensor):
        x = self.cn1(input_tensor)
        x = self.cn2(x)
        x = self.cn3(x)
        x = self.cn4(x)
        x = self.cn5(x)
        x = self.dense(x.view(-1, self.dense_size))
        return x



import numpy as np


def batch_split(input_tensor, batch_size):
    """Split an input tensor of size N x (dims) to a list of tensors of size batch_size x (dims).
    This function works with both numpy and torch tensors provided that dim 0 is the sampling dimension
    """

    return np.array_split(input_tensor, np.arange(batch_size, len(input_tensor), batch_size))





import os
import argparse
from concurrent.futures import ProcessPoolExecutor

import cv2
import numpy as np

from pytorchDL.tasks.image_segmentation.predictor import Predictor
from pytorchDL.utils.io import batch_split
from pytorchDL.utils.misc import generate_random_colors
from pytorchDL.loggers import ProgressLogger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--ckptPath', type=str, required=True, help='Path to checkpoint to be used for inference')
    parser.add_argument('--device', type=str, required=False, default='cpu', choices=['cpu', 'gpu'],
                        help='Device in which to run inference')
    parser.add_argument('--batchSize', type=int, required=False, default=1, help='Batch size')
    parser.add_argument('-i', '--input', required=True, type=str,
                        help='Accepted options: txt file with image paths, a directory containing images, an image file or "live" for real time inference on camera stream')
    parser.add_argument('-o', '--outDir', type=str, required=False,
                        help='Output directory to save prediction results. Not applicable of "live" input')
    parser.add_argument('-n', '--numProc', type=int, required=False, default=4,
                        help='Number of parallel processes for image loading')

    return parser.parse_args()


def main():

    args = parse_args()
    predictor = Predictor(args.ckptPath, device=args.device, num_proc=args.numProc)
    cl_colors = np.uint8(255 * generate_random_colors(n_colors=predictor.cfg['num_out_classes'], random_seed=42))

    if args.input.lower() == 'live':  # real time predictions from camera
        v_cap = cv2.VideoCapture(0)
        while v_cap.isOpened():
            ret, frame = v_cap.read()
            pred_seg = predictor.run(input=[frame])
            pred_seg_rgb = cl_colors[pred_seg]

            cv2.imshow('Inference', pred_seg_rgb)
            key = cv2.waitKey(25)

            if key == 27:  # press esc to stop program execution
                print('Execution manually terminated by user!!')
                v_cap.release()

    else:
        assert args.outDir, 'Output directory must be specified when input is not "live"'
        os.makedirs(args.outDir, exist_ok=True)

        if os.path.isfile(args.input):
            if '.txt' in args.input:  # input is a list of image paths
                with open(args.input, 'r') as fp:
                    input_img_paths = fp.read().splitlines()
            else:  # input is an image path
                input_img_paths = [args.input]

        elif os.path.isdir(args.input):  # input is a directory containing images
            input_img_paths = [os.path.join(args.input, x) for x in os.listdir(args.input)]

        input_batches = batch_split(input_img_paths, args.batchSize)
        prog_logger = ProgressLogger(total_steps=len(input_batches), description='Inference -- batch progress')

        with ProcessPoolExecutor(max_workers=args.numProc) as executor:
            for batch_paths in input_batches:
                batch_imgs = list(executor.map(cv2.imread, batch_paths))
                pred_seg_masks = predictor.run(input=batch_imgs)
                for pred_seg, in_path in zip(pred_seg_masks, batch_paths):
                    img_name = os.path.basename(in_path)
                    ext = '.' + img_name.split('.')[-1]

                    out_seg_path = os.path.join(args.outDir, img_name.replace(ext, '_infer' + ext))
                    cv2.imwrite(out_seg_path, pred_seg)

                    pred_seg_rgb = cl_colors[pred_seg]
                    out_seg_rgb_path = os.path.join(args.outDir, img_name.replace(ext, '_infer_rgb' + ext))
                    cv2.imwrite(out_seg_rgb_path, pred_seg_rgb)

                prog_logger.log()

        prog_logger.close()


if __name__ == '__main__':
    main()



import torch
from torchvision.models import vgg16


class FeatureExtractor(torch.nn.Module):

    def __init__(self, device='cpu'):
        super(FeatureExtractor, self).__init__()

        if device is None:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = torch.device('cuda' if device == 'gpu' else 'cpu')

        self.content_cn_idx = [3]
        self.style_cn_idx = [0, 1, 2, 3, 4]
        vgg = vgg16(pretrained=True).features.to(self.device).eval()
        self.model = vgg[0:11]   # extract up to conv_4 layer

        self.mean_rgb = torch.tensor([0.485, 0.456, 0.406]).to(self.device).view(-1, 1, 1)
        self.std_rgb = torch.tensor([0.229, 0.224, 0.225]).to(self.device).view(-1, 1, 1)

    def forward(self, x):
        cn_idx = 0
        content_features = []
        style_features = []

        # normalize input
        x = (x - self.mean_rgb) / self.std_rgb

        for layer in self.model:
            x = layer(x)
            if isinstance(layer, torch.nn.Conv2d):
                if cn_idx in self.content_cn_idx:
                    content_features.append(x)
                if cn_idx in self.style_cn_idx:
                    style_features.append(x)
                cn_idx += 1

        return content_features, style_features



import numpy as np
import cv2


def normalize(img, min_val=0.0, max_val=1.0):
    img = img.astype(float)
    return (img - img.min()) * (max_val - min_val) / (img.max() - img.min()) + min_val


def normalize_per_channel(img, min_val=0.0, max_val=1.0):
    ch = img.shape[-1]
    norm_img = np.concatenate([normalize(img[..., i], min_val, max_val) for i in range(ch)], axis=-1)
    return norm_img


def standardize(img):
    return (img - img.mean()) / img.std()


def standardize_per_channel(img):
    ch = img.shape[-1]
    norm_img = np.concatenate([standardize(img[..., i]) for i in range(ch)], axis=-1)
    return norm_img


def transform_image(image, rot_angle, x_shift=0, y_shift=0):
    h, w = image.shape[0:2]
    sx = w * x_shift
    sy = h * y_shift
    M = cv2.getRotationMatrix2D((w // 2, h // 2), -rot_angle, 1.0)
    M[0, 2] += sx
    M[1, 2] += sy
    trf_img = cv2.warpAffine(image, M, dsize=(w, h))
    return trf_img, M



import os
import torch
from pytorchDL.layers import ResidualConv2dLayer, FullyConnectedLayer, Conv2dLayer


def get_script_path():
    return os.path.realpath(__file__)


class ResidualBlock(torch.nn.Module):
    def __init__(self, in_features, n_res_units, bn=True):
        super(ResidualBlock, self).__init__()

        self.out_features = in_features

        res_units = list()
        res_units.append(ResidualConv2dLayer(in_features, activation='relu', bn=bn))
        for _ in range(1, n_res_units):
            res_units.append(ResidualConv2dLayer(in_features, activation='relu', bn=bn))

        self._res_block = torch.nn.Sequential(*res_units)

    def forward(self, input_tensor):
        return self._res_block(input_tensor)


class ResNet(torch.nn.Module):
    def __init__(self, input_size, num_out_classes, res_units_per_block=2, **kwargs):
        super().__init__()

        self.cn0 = Conv2dLayer(input_size[-1], 32, activation='relu')

        self.cn1 = Conv2dLayer(32, 32, activation='relu', stride=2)
        self.res_block1 = ResidualBlock(32, n_res_units=res_units_per_block)

        self.cn2 = Conv2dLayer(32, 64, activation='relu')
        self.res_block2 = ResidualBlock(64, n_res_units=res_units_per_block)

        self.cn3 = Conv2dLayer(64, 128, activation='relu', stride=2)
        self.res_block3 = ResidualBlock(128, n_res_units=res_units_per_block)

        self.cn4 = Conv2dLayer(128, 128, activation='relu')
        self.res_block4 = ResidualBlock(128, n_res_units=res_units_per_block)

        self.cn5 = Conv2dLayer(128, 256, activation='relu', stride=2)
        self.res_block5 = ResidualBlock(256, n_res_units=res_units_per_block)

        self.cn6 = Conv2dLayer(256, 512, activation='relu')

        self.fc1 = FullyConnectedLayer(512, num_out_classes, activation=None)

    def forward(self, input):
        x = self.cn0(input)

        x = self.cn1(x)
        x = self.res_block1(x)

        x = self.cn2(x)
        x = self.res_block2(x)

        x = self.cn3(x)
        x = self.res_block3(x)

        x = self.cn4(x)
        x = self.res_block4(x)

        x = self.cn5(x)
        x = self.res_block5(x)

        x = self.cn6(x)

        # global average pooling over channel dimension
        x = torch.mean(x, dim=[2, 3])

        out = self.fc1(x)
        return out



import cv2
import torch
import numpy as np

from pytorchDL.tasks.neural_style_transfer.feature_extractor import FeatureExtractor
from pytorchDL.utils.color import rgb_to_yuv, yuv_to_rgb
from pytorchDL.loggers import ProgressLogger


def compute_gram_matrix(input_tensor):
    a, b, c, d = input_tensor.size()

    features = input_tensor.view(a * b, c * d)
    gram_mat = torch.mm(features, features.t())  # compute the gram product
    return gram_mat / (a * b * c * d)


class NeuralStyleTransfer:
    """ Implementation of neural style transfer described in the paper by Gatys et al.
    'A Neural Algorithm of Artistic Style' https://arxiv.org/abs/1508.06576
    """

    def __init__(self, device=None, initialization_mode='content_image', color_mode='content_image', lr=0.05, content_weight=1.0, style_weight=1e05, tv_weight=1e-06):

        if device is None:
            self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.device = torch.device('cuda' if device == 'gpu' else 'cpu')

        self.feature_extractor = FeatureExtractor(device=device)
        if initialization_mode not in ['content_image', 'random']:
            raise Exception('Input initialization mode not valid. Options: "content_image", "random"')

        if color_mode not in ['content_image', 'free']:
            raise Exception('Input color mode not valid. Options: "preserve", "free"')

        self.cfg = {'init_mode': initialization_mode,
                    'color_mode': color_mode,
                    'lr': lr,
                    'content_weight': content_weight,
                    'style_weight': style_weight,
                    'total_variation_weight': tv_weight}

        self.output_result = None

    def _preproc_inputs(self, content_img, style_img, output_shape=None):
        if output_shape is None:
            out_h, out_w = content_img.shape[0:2]
        else:
            out_h, out_w = output_shape[0:2]
            content_img = cv2.resize(content_img, (out_w, out_h), interpolation=cv2.INTER_CUBIC)

        content_img = content_img[..., ::-1].astype(np.float32) / 255.0
        content_yuv = rgb_to_yuv(content_img)

        style_img = cv2.resize(style_img, (out_w, out_h), interpolation=cv2.INTER_CUBIC)
        style_img = style_img[..., ::-1].astype(np.float32) / 255.0

        content_tensor = torch.from_numpy(content_img[None]).to(self.device).permute(0, 3, 1, 2)
        style_tensor = torch.from_numpy(style_img[None]).to(self.device).permute(0, 3, 1, 2)

        return content_tensor, style_tensor, content_yuv

    def compute_losses(self, output_tensor, target_content_features, target_style_features):
        content_loss = torch.zeros(1, device=self.device)
        style_loss = torch.zeros(1, device=self.device)

        out_content_features, out_style_features = self.feature_extractor(output_tensor)

        n_content = len(out_content_features)
        n_style = len(out_style_features)

        # content loss
        for out_cf, target_cf in zip(out_content_features, target_content_features):
            content_loss += torch.nn.functional.mse_loss(out_cf, target_cf)
        content_loss /= n_content

        # style_loss
        for out_sf, target_sf in zip(out_style_features, target_style_features):
            out_gram = compute_gram_matrix(out_sf)
            target_gram = compute_gram_matrix(target_sf)
            style_loss += torch.nn.functional.mse_loss(out_gram, target_gram)
        style_loss /= n_style

        # total variation regularization
        tv_loss = torch.sum(torch.abs(output_tensor[..., :-1] - output_tensor[..., 1:])) + \
                  torch.sum(torch.abs(output_tensor[:, :, :-1, :] - output_tensor[:, :, 1:, :]))

        content_loss *= self.cfg['content_weight']
        style_loss *= self.cfg['style_weight']
        tv_loss *= self.cfg['total_variation_weight']
        total_loss = content_loss + style_loss + tv_loss
        return total_loss, content_loss, style_loss, tv_loss

    def run(self, content_img, style_img, steps, output_shape=None):

        cv2.imshow('Content image', content_img)
        cv2.imshow('Style image', style_img)

        content_tensor, style_tensor, content_yuv = self._preproc_inputs(content_img, style_img, output_shape)

        target_content_features, _ = self.feature_extractor(content_tensor)
        _, target_style_features = self.feature_extractor(style_tensor)

        target_content_features = [x.detach() for x in target_content_features]
        target_style_features = [x.detach() for x in target_style_features]

        if self.cfg['init_mode'] == 'content_image':
            output_tensor = content_tensor.clone()
        else:
            output_tensor = torch.randn(content_tensor.size(), device=self.device)

        output_tensor.requires_grad_()
        optimizer = torch.optim.Adam([output_tensor], lr=self.cfg['lr'])

        prog_bar = ProgressLogger(total_steps=steps, description='Neural style transfer')
        prog_bar.prepend_text('\nConfig:\n%s\n\n' % self.cfg)
        for step in range(steps):

            optimizer.zero_grad()

            total_loss, c_loss, s_loss, tv_loss = self.compute_losses(output_tensor,
                                                                      target_content_features,
                                                                      target_style_features)

            total_loss.backward()
            optimizer.step()

            prog_bar.log(total_loss=total_loss.item(), content_loss=c_loss.item(),
                         style_loss=s_loss.item(), tv_loss=tv_loss.item())

            # display current results
            out_img = output_tensor.detach().cpu().permute(0, 2, 3, 1).numpy()[0]
            out_img = np.clip(out_img, 0.0, 1.0)

            if self.cfg['color_mode'] == 'content_image':
                out_img_yuv = rgb_to_yuv(out_img)
                out_img = np.stack((out_img_yuv[..., 0], content_yuv[..., 1], content_yuv[..., 2]), axis=2)
                out_img = yuv_to_rgb(out_img)

            out_img = out_img[..., ::-1] * 255.0
            self.output_result = out_img.astype(np.uint8)
            cv2.imshow('Output image', self.output_result)
            cv2.waitKey(20)

        cv2.destroyAllWindows()



import os
import unittest

import numpy as np
from pytorchDL.metrics import ConfusionMatrix, MeanMetric


class TestMeanMetric(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.mean_metric = MeanMetric()

    def test_result(self):
        self.mean_metric(0)
        self.mean_metric(1)
        self.assertTrue(self.mean_metric.result() == 0.5)

        self.mean_metric(0)
        self.mean_metric(1)
        self.assertTrue(self.mean_metric.result() == 0.5)


class TestConfusionMatrix(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.num_classes = np.random.randint(2, 10)
        cls.conf_mat = ConfusionMatrix(num_classes=cls.num_classes)

    def test_update(self):
        preds = np.array([0, 0, 1, 1, 0, 0, 1])
        truth = np.array([0, 0, 1, 1, 1, 1, 0])
        cm_test = np.zeros((self.num_classes, self.num_classes))
        cm_test[0, 0] = 2
        cm_test[1, 1] = 2
        cm_test[1, 0] = 2
        cm_test[0, 1] = 1

        self.conf_mat.update(truth, preds)
        self.assertTrue(np.all(self.conf_mat.cm == cm_test))

        self.conf_mat.update(truth, preds)
        self.assertTrue(np.all(self.conf_mat.cm == 2*cm_test))
        self.assertTrue(np.sum(self.conf_mat.cm) == len(preds)*2)
        self.conf_mat.reset()

    def test_perfect_predictions(self):

        preds = np.arange(0, self.num_classes)
        truth = preds

        self.conf_mat.update(truth, preds)
        self.conf_mat.update(truth, preds)

        norm_cm = self.conf_mat.get_normalized()

        self.assertTrue(np.all(norm_cm == np.eye(self.num_classes)))
        self.conf_mat.reset()

    def test_wrong_predictions(self):

        preds = np.random.randint(1, self.num_classes, 10)
        truth = preds - 1

        self.conf_mat.update(truth, preds)
        self.conf_mat.update(truth, preds)

        norm_cm = self.conf_mat.get_normalized()
        self.assertTrue(np.all(np.diag(norm_cm) == np.zeros(self.num_classes)))
        self.conf_mat.reset()

    def test_plot_to_file(self):

        preds = np.arange(0, self.num_classes)
        truth = preds

        self.conf_mat.update(truth, preds)
        self.conf_mat.update(truth, preds)

        self.conf_mat.plot(to_file='test_conf_mat.png')
        self.assertTrue(os.path.exists('test_conf_mat.png'))
        os.remove('test_conf_mat.png')
        self.conf_mat.reset()



import os
from multiprocessing import cpu_count

import torch
import numpy as np

from pytorchDL.trainer_base import TrainerBase
from pytorchDL.loggers import TensorboardLogger, ProgressLogger
from pytorchDL.dataset_iterator import DataIterator
from pytorchDL.networks.unet import UNet
from pytorchDL.tasks.image_segmentation.data import Dataset
from pytorchDL.utils.misc import generate_random_colors

from pytorchDL.metrics import MeanMetric


class Trainer(TrainerBase):

    def __init__(self, trainer_mode, out_dir, batch_size, max_epochs, train_steps_per_epoch, val_steps_per_epoch, log_interval,
                 init_lr, **cfg_args):

        super().__init__(trainer_mode, out_dir, batch_size, max_epochs, train_steps_per_epoch, val_steps_per_epoch, log_interval,
                         **cfg_args)

        class_weights = torch.Tensor(self.cfg['class_weights']).cuda()
        self.set_model(model=UNet(input_channels=self.cfg['input_size'][-1],
                                  output_channels=self.cfg['num_out_classes']),
                       device=torch.device('cuda'))

        self.set_optimizer(optimizer=torch.optim.Adam(params=self.model.parameters(), lr=init_lr))
        self.set_loss_fn(torch.nn.CrossEntropyLoss(weight=class_weights), device=torch.device('cuda'))

        self.setup()
        self._set_label_colors()
        self.save_config(os.path.join(self.out_dir, 'trainer_cfg.json'))

    def _set_label_colors(self):
        self.label_colors = generate_random_colors(self.cfg['num_out_classes'], random_seed=42).T

    def _proc_output_for_log(self, batch_data, y_pred):
        x, y = batch_data
        x = x.cpu().detach().numpy()[0]
        y = y.cpu().detach().numpy()[0]

        y_pred = torch.nn.functional.softmax(y_pred, dim=1)
        y_pred = y_pred.cpu().detach().numpy()[0]
        pred_labels = np.argmax(y_pred, axis=0)

        gt_label_img = self.label_colors[:, y]
        gt_label_img = gt_label_img[None]

        pred_label_img = self.label_colors[:, pred_labels]
        pred_label_img = pred_label_img[None]

        x = x[None]
        return x, gt_label_img, pred_label_img

    def train_on_batch(self, batch_data):
        self.optimizer.zero_grad()

        # forward pass
        x, y = batch_data
        y_pred = self.model(x.cuda())
        batch_loss = self.loss_fn(y_pred, y.cuda())

        # backward pass
        batch_loss.backward()
        self.optimizer.step()

        # logging
        batch_loss = batch_loss.item()
        self.ep_train_mean_loss(batch_loss)  # update mean epoch loss metric
        self.prog_logger.log(batch_loss=batch_loss, mean_loss=self.ep_train_mean_loss.result())

        if (self.state['train_step'] % self.log_interval) == 0:
            x, gt, pred = self._proc_output_for_log(batch_data, y_pred)
            log_data = [{'data': x, 'type': 'image', 'name': '0_input'},
                        {'data': gt, 'type': 'image', 'name': '1_gt_mask'},
                        {'data': pred, 'type': 'image', 'name': '2_pred_mask'},
                        {'data': batch_loss, 'type': 'scalar', 'name': 'batch_loss'}]

            self.tb_logger.log(log_data, stage='train', step=self.state['train_step'])

        self.state['train_step'] += 1  # update train step

    def eval_on_batch(self, batch_data):

        # forward pass
        x, y = batch_data
        y_pred = self.model(x.cuda())
        batch_loss = self.loss_fn(y_pred, y.cuda())

        # logging
        batch_loss = batch_loss.item()
        self.ep_val_mean_loss(batch_loss)  # update mean epoch loss metric
        self.prog_logger.log(batch_loss=batch_loss, mean_loss=self.ep_val_mean_loss.result())

        if (self.state['val_step'] % self.log_interval) == 0:
            x, gt, pred = self._proc_output_for_log(batch_data, y_pred)
            log_data = [
                {'data': x, 'type': 'image', 'name': '0_input'},
                {'data': gt, 'type': 'image', 'name': '1_gt_mask'},
                {'data': pred, 'type': 'image', 'name': '2_pred_mask'}]
            self.tb_logger.log(log_data, stage='val', step=self.state['val_step'])

        self.state['val_step'] += 1

    def _init_metrics(self):
        self.ep_train_mean_loss = MeanMetric()
        self.ep_val_mean_loss = MeanMetric()

    def _reset_metrics(self):
        self.ep_train_mean_loss.reset()
        self.ep_val_mean_loss.reset()

    def run(self):

        num_dataloader_workers = 0 if self.trainer_mode == 'debug' else cpu_count() - 2

        # load train and validation datasets iterators
        train_dataset = Dataset(data_dir=self.cfg['train_data_dir'], output_shape=self.cfg['input_size'])
        train_data_iterator = DataIterator(train_dataset, batch_size=self.batch_size,
                                           num_workers=num_dataloader_workers, shuffle=True)

        val_dataset = Dataset(data_dir=self.cfg['val_data_dir'], output_shape=self.cfg['input_size'])
        val_data_iterator = DataIterator(val_dataset, batch_size=self.batch_size,
                                         num_workers=num_dataloader_workers, shuffle=True)

        if self.train_steps_per_epoch <= 0:  # if train steps per epoch is <= 0, set it to cover the whole dataset
            self.train_steps_per_epoch = len(train_dataset) // self.batch_size

        if self.val_steps_per_epoch <= 0:
            self.val_steps_per_epoch = len(val_dataset) // self.batch_size

        self.tb_logger = TensorboardLogger(log_dir=os.path.join(self.log_dir, 'tensorboard'))
        self._init_metrics()

        for ep in range(self.state['epoch'], self.max_epochs):
            print('\nEPOCH: %d' % ep)
            self.state['epoch'] = ep

            self._reset_metrics()

            # TRAIN LOOP
            self.model.train()
            self.stage = 'train'
            self.prog_logger = ProgressLogger(total_steps=self.train_steps_per_epoch, description='Training')

            for i in range(self.train_steps_per_epoch):

                batch_data = next(train_data_iterator)
                self.train_on_batch(batch_data)

            self.tb_logger.log(log_data=[{'data': self.ep_train_mean_loss.result(), 'type': 'scalar',
                                          'name': '%s/ep_mean_loss' % self.stage, 'stage': self.stage}],
                               stage=self.stage, step=ep)

            self.prog_logger.close()
            self.save_checkpoint('checkpoint-step-%d' % self.state['train_step'])

            # VAL LOOP
            self.model.eval()
            self.stage = 'val'
            self.prog_logger = ProgressLogger(total_steps=self.val_steps_per_epoch, description='Validation')

            with torch.no_grad():
                for i in range(self.val_steps_per_epoch):

                    batch_data = next(val_data_iterator)
                    self.eval_on_batch(batch_data)

            self.tb_logger.log(log_data=[{'data': self.ep_val_mean_loss.result(), 'type': 'scalar',
                                          'name': '%s/ep_mean_loss' % self.stage, 'stage': self.stage}],
                               stage=self.stage,
                               step=ep)

            self.prog_logger.close()

            if self.ep_val_mean_loss.result() < self.state['best_val_loss']:
                print('\tMean validation loss decreased from %f to %f. Saving best model' % (self.state['best_val_loss'], self.ep_val_mean_loss.result()))
                self.state['best_val_loss'] = self.ep_val_mean_loss.result()
                self.save_best_checkpoint()

    def run_testing(self):
        raise NotImplementedError()



import torch
import numpy as np
import cv2

from pytorchDL.networks.resnet import ResNet


class Predictor:

    def __init__(self, ckpt_path, device, num_proc=1):

        self.device = torch.device('cuda' if device == 'gpu' else 'cpu')
        checkpoint = torch.load(ckpt_path, map_location=self.device)

        self.cfg = checkpoint['trainer_cfg']

        self.model = ResNet(**self.cfg)
        self.model.load_state_dict(checkpoint['model_state'])
        self.model.to(self.device)
        self.model.eval()
        self.num_proc = num_proc

    def preprocess_single_image(self, img):

        h, w, ch = self.cfg['input_size']
        if ch == 1:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        else:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        ih, iw = img.shape[0:2]
        interp_method = cv2.INTER_AREA if (w < iw and h < ih) else cv2.INTER_CUBIC
        img = cv2.resize(img, (w, h), interpolation=interp_method)  # resize to network training input size
        img = img.astype(np.float32) / 255.0

        if ch == 1:
            img = img[..., None]

        return img

    def preprocess_input(self, images_bgr):

        proc_images = [self.preprocess_single_image(x) for x in images_bgr]
        img_tensor = torch.tensor(proc_images).permute(0, 3, 1, 2).to(self.device)
        return img_tensor

    def predict_on_batch(self, batch):

        pred_logits = self.model(batch)
        label_probs = torch.nn.functional.softmax(pred_logits, dim=1)
        pred_label_prob, pred_labels = torch.max(label_probs, dim=1)

        pred_labels = pred_labels.cpu().numpy()
        pred_label_prob = pred_label_prob.cpu().numpy()
        return pred_labels, pred_label_prob

    def run(self, input):
        """Runs class inference over input images

        :param input: list of BGR image arrays
        :return: a N x 2 numpy array containing the predicted label and probability for each input in each row
        """

        with torch.no_grad():
            input_tensor = self.preprocess_input(input)
            p_labels, p_probs = self.predict_on_batch(input_tensor)
        predictions = np.stack((p_labels, p_probs), axis=1)

        return predictions



import argparse
import traceback

from pytorchDL.tasks.image_segmentation.trainer import Trainer


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', type=str, required=True, choices=['start', 'resume', 'test', 'debug'], help='Trainer mode')
    parser.add_argument('--trainDir', type=str, required=True, help='Folder containing train images and labels')
    parser.add_argument('--valDir', type=str, required=True, help='Folder containing val images and labels')
    parser.add_argument('--outDir', type=str, required=True, help='Output folder')

    parser.add_argument('--batchSize', type=int, required=True, help='Training batch size')
    parser.add_argument('--numClasses', type=int, required=True, help='Number of classes in data')
    parser.add_argument('--imgSize', type=str, required=True, help='Input training image size -> H,W,Ch  e.g "256,256,1" ')

    parser.add_argument('--maxEpochs', type=int, required=False, default=100, help='Maximum number of epochs to train')
    parser.add_argument('--testDir', type=str, required=False, help='Folder containing test images and labels')

    parser.add_argument('--trainStepsPerEpoch', type=int, required=False, default=-1, help='Number of steps that define one epoch. Default = -1 -->  all dataset examples')
    parser.add_argument('--logInterval', type=int, required=False, default=15, help='Log step interval. Default = log every 15 steps')
    parser.add_argument('--initLR', type=float, required=False, default=0.001)
    parser.add_argument('--classWeights', type=str, required=False, help='Weight for each class, as a comma separated strin of floats e.g "1,2,0.5,7"')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    input_size = list(map(int, args.imgSize.split(',')))

    if args.classWeights is not None:
        class_weights = tuple(map(float, args.imgSize.split(',')))
        assert len(class_weights) == args.numClasses
    else:
        class_weights = tuple([1.0] * args.numClasses)

    cfg = {'train_data_dir': args.trainDir,
           'val_data_dir': args.valDir,
           'input_size': input_size,
           'num_out_classes': args.numClasses,
           'class_weights': class_weights}

    trainer = Trainer(trainer_mode=args.mode,
                      out_dir=args.outDir,
                      batch_size=args.batchSize,
                      max_epochs=args.maxEpochs,
                      train_steps_per_epoch=args.trainStepsPerEpoch,
                      val_steps_per_epoch=-1,
                      log_interval=args.logInterval,
                      init_lr=args.initLR,
                      **cfg)

    try:
        trainer.run()
    except KeyboardInterrupt:
        print('\n\nExecution terminated manually by the user!!! Saving checkpoint...\n\n')
        trainer.save_checkpoint('checkpoint-step-%d-interrupt' % trainer.state['train_step'])
    except Exception as error:
        print('\n\n')
        traceback.print_exc()


if __name__ == '__main__':
    main()



import os

import torch
import numpy as np

from torch.utils.tensorboard import SummaryWriter
from pytorchDL.utils.misc import get_current_time
from tensorboard import program
from tqdm import trange


class TensorboardLogger:

    def __init__(self, log_dir, port=3468):
        self.summary_dir = os.path.join(log_dir, get_current_time())
        self.summary_writer = SummaryWriter(self.summary_dir, flush_secs=15)
        self._launch_tensorboard(port=port)

    def _launch_tensorboard(self, port):
        """
        Launch tensorboard in background to locally inspect the created summary under a specified port
        """
        self._tb = program.TensorBoard()
        self._tb.configure(argv=[None, '--logdir', os.path.dirname(self.summary_dir), '--port', str(port)])
        self.tensorboard_url = self._tb.launch()
        self.print_tensorboard_url()

    def print_tensorboard_url(self):
        print('\n\nTensorboard url: %s\n\n' % self.tensorboard_url)

    def log(self, log_data, stage, step):
        """
        Log a list of data to tensorboard. Step is automatically determined from the trainer current state.
        Each piece of data to be logged must be defined as a dict, with fields:
            'type': type of data ('scalar' or 'image')
            'name': the name of the log in which this new data will be included (e.g 'batch_loss', 'accuracy')
            'data': numpy array or torch tensor representing the data.
                    If image data, use NCHW format, float type and intensity range between [0, 1]
                    If pointcloud data, use Nx6xP format, where P is the number of points and the first dimension represents [x y z R G B]
                        RGB values must be in the range [0, 1]

        :param log_data: list of dicts, each one containing a piece of data to be logged. This dict must have 'type', 'name' and 'data' fields
        :param stage: string naming the training stage for the data to log. (e.g 'train', 'val', 'test')
        """

        for data_dict in log_data:

            tag = '%s/%s' % (stage, data_dict['name'])
            if data_dict['type'] == 'scalar':
                self.summary_writer.add_scalar(tag=tag, scalar_value=data_dict['data'], global_step=step)
            elif data_dict['type'] == 'image':
                self.summary_writer.add_images(tag=tag, img_tensor=data_dict['data'], global_step=step)
            elif data_dict['type'] == 'pointcloud':
                if isinstance(data_dict['data'], np.ndarray):
                    data_dict['data'] = torch.from_numpy(data_dict['data'])
                vertices = data_dict['data'][:, 0:3, :].permute(0, 2, 1)

                colors = 255 * data_dict['data'][:, 3:6, :].permute(0, 2, 1)
                colors = colors.type(torch.uint8)
                self.summary_writer.add_mesh(tag=tag, vertices=vertices, colors=colors, global_step=step)
            else:
                raise Exception('Logging input data type (%s) is not implemented' % data_dict['type'])


class ProgressLogger:

    def __init__(self, total_steps, description='Progress'):
        self.prog_bar = trange(total_steps, desc=description)

    def log(self, **kwargs):
        self.prog_bar.set_postfix(kwargs)
        self.prog_bar.update()

    def prepend_text(self, text):
        self.prog_bar.write(text)

    def close(self):
        self.prog_bar.close()



import os
import glob

import cv2
import torch
import numpy as np


class Dataset(torch.utils.data.Dataset):

    def __init__(self, data_dir, output_shape):
        """Data directory must an "images" folder with the images and a "masks" folder with the corresponding label
         masks. Corresponding image and mask files must share the same filename and extension.

        :param data_dirs: directory containing the image and label data.
        :param output_shape: list or tuple defining the generator output shape
        """

        self.data_files = []
        img_dir = os.path.join(data_dir, 'images')
        mask_dir = os.path.join(data_dir, 'masks')
        img_files = glob.glob(os.path.join(img_dir, '*'))
        for img_path in img_files:
            mask_path = img_path.replace(img_dir, mask_dir)
            if os.path.exists(mask_path):
                self.data_files.append([img_path, mask_path])

        self.shuffle()
        self.output_shape = output_shape

    def __len__(self):
        return len(self.data_files)

    def __getitem__(self, index):

        h, w, ch = self.output_shape
        if ch > 1:
            img = cv2.imread(self.data_files[index][0])
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        else:
            img = cv2.imread(self.data_files[index][0], 0)
            img = np.expand_dims(img, axis=2)

        labels = cv2.imread(self.data_files[index][1], 0)

        img = cv2.resize(img, (w, h))
        labels = cv2.resize(labels, (w, h), interpolation=cv2.INTER_NEAREST)

        img = img.astype(np.float32) / 255

        if ch == 1:
            x = torch.from_numpy(img[None])
        else:
            x = torch.from_numpy(img).permute(dims=(2, 0, 1))

        y = torch.tensor(labels).type(torch.long)
        return x, y

    def shuffle(self):
        self.data_files = np.random.permutation(self.data_files)



import numpy as np


class MeanMetric:

    def __init__(self):
        self.acc = 0
        self.it = 0

    def __call__(self, value):
        self.acc += value
        self.it += 1

    def reset(self):
        self.acc = 0
        self.it = 0

    def result(self):
        return self.acc / self.it


class ConfusionMatrix:

    def __init__(self, num_classes, tags=None):
        self.num_classes = num_classes
        self.tags = tags if tags is not None else ['class_%d' % i for i in range(num_classes)]
        self.cm = np.zeros((num_classes, num_classes))

    def reset(self):
        self.cm = np.zeros((self.num_classes, self.num_classes))

    def update(self, gt_labels, pred_labels):
        for gt, pred in zip(gt_labels, pred_labels):
            self.cm[gt, pred] += 1

    def get_normalized(self):
        norm_conf_mat = self.cm / np.sum(self.cm, axis=1)[:, None]
        norm_conf_mat[np.isnan(norm_conf_mat)] = 0
        return norm_conf_mat

    def plot(self, title='Confusion Matrix', normalized=True, to_file=None):
        import matplotlib.pyplot as plt
        import seaborn as sns

        if normalized:
            cm = self.get_normalized()
        else:
            cm = self.cm

        fmt = 'd' if np.issubdtype(cm.dtype, np.integer) else '.3f'
        plt.figure(figsize=(15, 15))
        ax = sns.heatmap(cm, annot=True, xticklabels=self.tags, yticklabels=self.tags, fmt=fmt, cmap='Blues')
        ax.set(xlabel='PREDICTED', ylabel='GROUND TRUTH')
        plt.title(title)
        if to_file is None:
            plt.show()
        else:
            plt.savefig(to_file)

        return ax


class ConfusionMatrixTopK(ConfusionMatrix):

    def __init__(self, num_classes, k, tags=None):
        super().__init__(num_classes=num_classes, tags=tags)
        self.k = k

    def update(self, gt_labels, pred_probs):

        topk = np.argsort(-pred_probs, axis=1)[:, 0:self.k]
        for gt, top_preds in zip(gt_labels, topk):
            if gt in top_preds:
                self.cm[gt, gt] += 1
            else:
                self.cm[gt, top_preds[0]] += 1



import os
import argparse
from concurrent.futures import ProcessPoolExecutor

import cv2
import numpy as np
import pandas as pd

from pytorchDL.tasks.image_classification.predictor import Predictor
from pytorchDL.utils.io import batch_split
from pytorchDL.loggers import ProgressLogger


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--ckptPath', type=str, required=True, help='Path to checkpoint to be used for inference')
    parser.add_argument('--device', type=str, required=False, default='cpu', choices=['cpu', 'gpu'],
                        help='Device in which to run inference')
    parser.add_argument('--batchSize', type=int, required=False, default=1, help='Batch size for class inference')
    parser.add_argument('-i', '--input', required=True, type=str,
                        help='Accepted options: txt file with image paths, a directory containing images, an image file or "live" for real time inference on camera stream')
    parser.add_argument('-o', '--output', type=str, required=False,
                        help='Output CSV file to save prediction results. Not applicable of "live" input')
    parser.add_argument('-n', '--numProc', type=int, required=False, default=4,
                        help='Number of parallel processes for image loading')

    return parser.parse_args()


def main():

    args = parse_args()
    predictor = Predictor(args.ckptPath, device=args.device, num_proc=args.numProc)

    if args.input.lower() == 'live':  # real time predictions from camera
        v_cap = cv2.VideoCapture(0)
        while v_cap.isOpened():
            ret, frame = v_cap.read()
            fh, fw = frame.shape[0:2]

            preds = predictor.run(input=[frame])
            label, prob = preds[0]

            pred_str = 'Pred label: %d -- Prob: %.3f' % (label, prob)
            cv2.putText(frame, pred_str, (10, fh - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 255), 1, cv2.LINE_AA)
            cv2.imshow('Inference', frame)
            key = cv2.waitKey(25)

            if key == 27:  # press esc to stop program execution
                print('Execution manually terminated by user!!')
                v_cap.release()

    else:
        assert args.output, 'Output file must be specified when input is not "live"'

        predictions = list()
        if os.path.isfile(args.input):
            if '.txt' in args.input:  # input is a list of image paths
                with open(args.input, 'r') as fp:
                    input_img_paths = fp.read().splitlines()
            else:  # input is an image path
                input_img_paths = [args.input]

        elif os.path.isdir(args.input):  # input is a directory containing images
            input_img_paths = [os.path.join(args.input, x) for x in os.listdir(args.input)]

        input_batches = batch_split(input_img_paths, args.batchSize)
        prog_logger = ProgressLogger(total_steps=len(input_batches), description='Inference -- batch progress')

        with ProcessPoolExecutor(max_workers=args.numProc) as executor:
            for batch_paths in input_batches:
                batch_imgs = list(executor.map(cv2.imread, batch_paths))
                preds = predictor.run(input=batch_imgs)
                predictions.append(preds)
                prog_logger.log()

        prog_logger.close()
        predictions = np.concatenate(predictions, axis=0)
        output_data = {'IMAGE': input_img_paths,
                       'PRED_CLASS_ID': predictions[:, 0],
                       'PRED_CLASS_PROB': predictions[:, 1]}

        # save output predictions to a csv file
        df = pd.DataFrame(data=output_data)
        df = df.astype({'PRED_CLASS_ID': int})

        os.makedirs(os.path.dirname(args.output), exist_ok=True)
        df.to_csv(args.output, index=False)


if __name__ == '__main__':
    main()



import os
import json

import torch
from tqdm import tqdm

from pytorchDL.tasks.image_classification.predictor import Predictor
from pytorchDL.tasks.image_classification.data import Dataset
from pytorchDL.metrics import ConfusionMatrix


class Evaluator(Predictor):

    def __init__(self, test_data_dir, out_dir, ckpt_path, batch_size, device, num_proc):
        super().__init__(ckpt_path, device, num_proc=num_proc)
        self.test_data_dir = test_data_dir
        self.out_dir = out_dir
        os.makedirs(out_dir, exist_ok=True)

        self.batch_size = batch_size

    def run_testing(self):
        test_dataset = Dataset(data_dir=self.test_data_dir, output_shape=self.cfg['input_size'])
        test_dataloader = torch.utils.data.DataLoader(dataset=test_dataset,
                                                      batch_size=self.batch_size,
                                                      num_workers=self.num_proc)

        cm = ConfusionMatrix(num_classes=self.cfg['num_out_classes'], tags=test_dataset.class_tags)
        test_steps = len(test_dataset) // self.batch_size
        test_results = {}
        with torch.no_grad():
            for batch_data in tqdm(test_dataloader, total=test_steps):

                x, y = batch_data
                x = x.to(self.device)

                pred_logits = self.model(x)
                pred_logits = torch.nn.functional.softmax(pred_logits, dim=1)
                _, pred_labels = pred_logits.max(dim=1)

                gt_labels = y.cpu().numpy()
                pred_labels = pred_labels.cpu().numpy()
                cm.update(gt_labels, pred_labels)

        out_file = os.path.join(self.out_dir, 'test_confusion_matrix.png')
        cm.plot(title='Conf. Matrix - Classification', normalized=True, to_file=out_file)

        test_results['norm_conf_mat'] = cm.get_normalized().tolist()
        test_results['class_tags'] = test_dataset.class_tags
        with open(os.path.join(self.out_dir, 'test_results.json'), 'w') as fp:
            json.dump(test_results, fp)


if __name__ == '__main__':

    test_dir = '/media/miguel/HDD/DeepLearning/Datasets/kaggle_mnist/val'
    out_dir = '/home/miguel/prueba_clasificacion_mnist/exp_0'
    ckpt_path = '/home/miguel/prueba_clasificacion_mnist/exp_0/checkpoints/best_checkpoint.pth'

    evaluator = Evaluator(test_data_dir=test_dir,
                          out_dir=out_dir,
                          ckpt_path=ckpt_path,
                          batch_size=32,
                          device='gpu',
                          num_proc=4)
    evaluator.run_testing()



from time import time
from datetime import datetime

import numpy as np


def print_in_place(text):
    print('\r' + text + '\t' * 5, end='', sep='')


class Timer:

    def __init__(self, total_steps):
        self.start_time = time()
        self.total_steps = total_steps
        self.it = 0

    def __call__(self):
        self.it += 1

        elapsed_time = time() - self.start_time
        est_time_per_step = elapsed_time / self.it
        time_left = est_time_per_step * (self.total_steps - self.it)
        m, s = divmod(time_left, 60)
        h, m = divmod(m, 60)
        time_left = "%d:%02d:%02d" % (h, m, s)
        return time_left, est_time_per_step


def get_current_time():
    return datetime.now().strftime('%Y-%m-%d_%H:%M:%S')


def generate_random_colors(n_colors, random_seed=None):
    if random_seed is not None:
        np.random.seed(random_seed)
    colors = np.random.uniform(0, 1, size=(n_colors, 3))
    if random_seed is not None:
        np.random.seed()

    return colors



import argparse
import traceback

import cv2
from pytorchDL.tasks.neural_style_transfer.predictor import NeuralStyleTransfer


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-c', '--contentImg', type=str, required=True, help='Content image path')
    parser.add_argument('-s', '--styleImg', type=str, required=True, help='Style image path')
    parser.add_argument('-o', '--outputPath', type=str, required=True, help='Output image path')
    parser.add_argument('--maxIter',  type=int, default=1500, required=False, help='Maximum number of iterations')

    parser.add_argument('--device', type=str, default=None, required=False, choices=['gpu', 'cpu'], help='Device selection')
    parser.add_argument('--outputSize', type=str, required=False, default=None, help='Output image size -> H,W  e.g "512,512". Default is content image size')
    parser.add_argument('--initMode', type=str, default='content_image', required=False, choices=['content_image', 'random'], help='Output image initialization')
    parser.add_argument('--colorMode', type=str, default='free', required=False, choices=['content_image', 'free'], help='Keep content image color or allow free color transfer')
    parser.add_argument('-lr', '--learningRate', type=float, default=0.05, required=False, help='Learning rate')
    parser.add_argument('--contentWeight', type=float, default=1.0, required=False, help='Content loss weight')
    parser.add_argument('--styleWeight', type=float, default=1e05, required=False, help='Style loss weight')
    parser.add_argument('--totalVariationWeight', type=float, default=1e-05, required=False, help='Total variation regularization weight')

    return parser.parse_args()


def main():

    args = parse_args()
    nst = NeuralStyleTransfer(device=args.device,
                              initialization_mode=args.initMode,
                              color_mode=args.colorMode,
                              lr=args.learningRate,
                              content_weight=args.contentWeight,
                              style_weight=args.styleWeight,
                              tv_weight=args.totalVariationWeight)

    content_img = cv2.imread(args.contentImg)
    style_img = cv2.imread(args.styleImg)

    output_shape = list(map(int, args.outputSize.split(','))) if args.outputSize is not None else None

    try:
        nst.run(content_img, style_img, steps=args.maxIter, output_shape=output_shape)
        print('\n\nMaximum steps reached. Saving output to: %s\n\n' % args.outputPath)
        cv2.imwrite(args.outputPath, nst.output_result)

    except KeyboardInterrupt:
        print('\n\nExecution terminated manually by the user!!! Saving current output to: %s\n\n' % args.outputPath)
        cv2.imwrite(args.outputPath, nst.output_result)

    except Exception as error:
        print('\n\n')
        traceback.print_exc()


if __name__ == '__main__':
    main()



import os
import glob
import random

import cv2
import torch
import numpy as np


class Dataset(torch.utils.data.Dataset):

    def __init__(self, data_dir, output_shape):
        """Data directory must contain one image directory per class, name after the class name.

        :param data_dir: directory containing the image and label data.
        :param output_shape: list or tuple defining the generator output shape
        """

        self.data_files = []
        self.class_tags = sorted(os.listdir(data_dir))

        for i, tag in enumerate(self.class_tags):
            class_dir = os.path.join(data_dir, tag)
            img_files = glob.glob(os.path.join(class_dir, '*'))
            for path in img_files:
                self.data_files.append([path, i])

        self.shuffle()
        self.output_shape = output_shape

    def __len__(self):
        return len(self.data_files)

    def __getitem__(self, index):
        h, w, ch = self.output_shape

        img_path, label = self.data_files[index]
        if ch > 1:
            img = cv2.imread(img_path)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        else:
            img = cv2.imread(img_path, 0)

        img = cv2.resize(img, (w, h))
        img = img.astype(np.float32) / 255

        if ch == 1:
            x = torch.from_numpy(img[None])
        else:
            x = torch.from_numpy(img).permute(dims=(2, 0, 1))
        y = torch.tensor(label)
        return x, y

    def shuffle(self):
        random.shuffle(self.data_files)



import os
import setuptools


def get_requirements():
    cwd = os.path.dirname(os.path.realpath(__file__))
    req_file = os.path.join(cwd, 'requirements.txt')
    with open(req_file, 'r') as fp:
        reqs = fp.read().splitlines()

    return reqs


def get_version():
    cwd = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(cwd, 'pytorchDL', '__init__.py'), 'r') as fp:
        lines = fp.read().splitlines()

    for line in lines:
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise Exception('Could not retrieve version number')


setuptools.setup(
    name="pytorchDL",
    version=get_version(),
    author="Milogav",
    description="Package containing network definitions and utilities for pytorch deep learning framework",
    url='https://github.com/Milogav/PytorchDL',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[get_requirements()],
    entry_points={
        "console_scripts": [
            "pdl-train-image-classifier = apps.image_classification.run_training:main",
            "pdl-test-image-classifier = apps.image_classification.run_testing:main",
            "pdl-predict-image-classifier = apps.image_classification.run_inference:main",
            "pdl-train-image-segmenter = apps.image_segmentation.run_training:main",
            "pdl-neural-style-transfer = apps.neural_style_transfer.run_neural_style_transfer:main"
        ]
    }
)



import os
import json

import torch
import numpy as np


class TrainerBase:

    def __init__(self, trainer_mode, out_dir, batch_size, max_epochs, train_steps_per_epoch, val_steps_per_epoch, log_interval, **cfg_args):

        """
        Setup the trainer attributes
        :param out_dir: main output directory where logs and checkpoints will be saved
        :param batch_size:
        :param max_epochs: maximum training epoch number
        :param train_steps_per_epoch: train steps per epoch
        :param val_steps_per_epoch: validation steps per epoch
        :param log_interval: train/val steps between logs
        :param cfg_args: any additional configuration. It will be stored in self.cfg as a dict
        :return:
        """

        self.out_dir = out_dir
        self.log_dir = os.path.join(out_dir, 'logs')
        self.checkpoint_dir = os.path.join(out_dir, 'checkpoints')
        self.batch_size = batch_size
        self.max_epochs = max_epochs
        self.train_steps_per_epoch = train_steps_per_epoch
        self.val_steps_per_epoch = val_steps_per_epoch
        self.log_interval = log_interval
        self.trainer_mode = trainer_mode

        os.makedirs(self.log_dir, exist_ok=True)
        os.makedirs(self.checkpoint_dir, exist_ok=True)

        self.state = {'epoch': 0,
                      'train_step': 0,
                      'val_step': 0,
                      'best_val_loss': np.inf}

        self.model = None
        self.optimizer = None
        self.loss_fn = None
        self.cfg = cfg_args

    def setup(self):

        if self.trainer_mode == 'start':
            if len(os.listdir(self.checkpoint_dir)) > 0:
                raise Exception('Error! Output checkpoint dir (%s) not empty, which is incompatible with "%s" trainer mode'
                                % (self.cfg['checkpoint_dir'], self.trainer_mode))
        elif self.trainer_mode == 'resume':
            if not len(os.listdir(self.checkpoint_dir)):
                raise Exception('Error! Cannot resume from an empty checkpoint dir. Use "start" trainer mode instead')
            self.load_last_checkpoint(self.checkpoint_dir)
        elif self.trainer_mode == 'test':
            if not len(os.listdir(self.checkpoint_dir)):
                raise Exception('Error! Cannot load best checkpoint from an empty checkpoint dir.')
            self.load_best_checkpoint(self.checkpoint_dir)
        elif self.trainer_mode == 'debug':
            pass
        else:
            raise Exception('Error! Input trainer mode (%s) not available' % self.trainer_mode)

    def set_model(self, model, device=None):
        self.model = model
        if device is not None:
            self.model.to(device)

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def set_loss_fn(self, loss_fn, device=None):
        self.loss_fn = loss_fn
        if device is not None:
            self.loss_fn.to(device)

    def set_config(self, **cfg_args):
        self.cfg.update(cfg_args)

    def save_config(self, out_path):
        """
        Export the trainer configuration to a json file
        """

        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        with open(out_path, 'w') as fp:
            json.dump(self.cfg, fp)

    def load_config(self, cfg_json):
        """
        Import a previously exported trainer configuration from a json file
        """

        with open(cfg_json, 'r') as fp:
            self.cfg = json.load(fp)

    def get_last_checkpoint(self):
        info_last_ckpt = os.path.join(self.cfg['checkpoint_dir'], 'last_checkpoint.txt')
        try:
            with open(info_last_ckpt, 'r') as fp:
                last_checkpoint_name = fp.read().strip()
        except FileNotFoundError:
            print('"last_checkpoint.txt" not found in checkpoint directory: %s' % self.cfg['checkpoint_dir'])
            last_checkpoint_name = None

        return last_checkpoint_name

    def save_checkpoint(self, name):
        """Saves a generic checkpoint dict to the output checkpoint directory. If working with custom checkpoints,
        this method must be overridden in the children class that inherits from TrainerBase

        :param name: output checkpoint filename without extension
        """

        # print('\tSaving checkpoint { %s } in %s' % (name, self.cfg['checkpoint_dir']))
        checkpoint = {
            'model_state': self.model.state_dict(),
            'optimizer_state': self.optimizer.state_dict(),
            'loss_fn_state': self.loss_fn.state_dict(),
            'trainer_state': self.state,
            'trainer_cfg': self.cfg
        }

        ckpt_path = os.path.join(self.checkpoint_dir, name+'.pth')
        torch.save(checkpoint, ckpt_path)

        info_last_ckpt = os.path.join(self.checkpoint_dir, 'last_checkpoint.txt')
        with open(info_last_ckpt, 'w') as fp:
            fp.write(name + '.pth')

    def save_best_checkpoint(self):
        self.save_checkpoint('best_checkpoint')

    def load_checkpoint(self, ckpt_path):
        """Loads a generic checkpoint. If working with custom checkpoints, this method must be overridden
        in the children class that inherits from TrainerBase

        :param ckpt_path: path to the checkpoint file to be loaded
        """

        print('\tLoading checkpoint from: %s' % ckpt_path)
        checkpoint = torch.load(ckpt_path)

        self.state = checkpoint['trainer_state']
        self.loss_fn.load_state_dict(checkpoint['loss_fn_state'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state'])
        self.model.load_state_dict(checkpoint['model_state'])

    def load_last_checkpoint(self, ckpt_dir):

        print('\tLoading last checkpoint from folder: %s' % ckpt_dir)
        self.load_checkpoint(ckpt_path=os.path.join(ckpt_dir, self.get_last_checkpoint()))

    def load_best_checkpoint(self, ckpt_dir):
        print('\tLoading best checkpoint from folder: %s' % ckpt_dir)
        self.load_checkpoint(ckpt_path=os.path.join(ckpt_dir, 'best_checkpoint.pth'))



import torch


class DataIterator(object):

    def __init__(self, dataset, batch_size, num_workers, shuffle=True):
        """Core class for iterating over a pytorch dataset. This class handles proper shuffling application, batching and infinite
        iteration over the input dataset.

        :param dataset: a pytorch dataset object inheriting from pytorch Dataset class.
                        The dataset object must contain ".shuffle() method that shuffles data
        :param batch_size: integer that specifies the number of elements per batch
        :param num_workers: number of parallel dataloaders
        :param shuffle: boolean to specify if the data must be shuffled or not
        """

        self._dataset = dataset
        self._shuffle = shuffle
        if self._shuffle:
            self._dataset.shuffle()

        self._dataloader = torch.utils.data.DataLoader(self._dataset,
                                                       batch_size=batch_size,
                                                       shuffle=shuffle,
                                                       num_workers=num_workers)
        self.num_examples = len(self._dataset)
        self._dataset_iterator = iter(self._dataloader)

    def __len__(self):
        return self.num_examples

    def __getitem__(self, item):
        return self._dataset[item]

    def __next__(self):
        try:
            batch_data = next(self._dataset_iterator)
        except StopIteration:  # when dataset is exhausted, re shuffle (if applicable) it and reload iterator
            if self._shuffle:
                self._dataset.shuffle()
            self._dataset_iterator = iter(self._dataloader)
            batch_data = next(self._dataset_iterator)

        return batch_data



import torch
import os
from pytorchDL.layers import Conv2dLayer, FullyConnectedLayer


def get_script_path():
    return os.path.realpath(__file__)


class VGG11(torch.nn.Module):

    def __init__(self, input_size, num_out_classes, **kwargs):
        super(VGG11, self).__init__()

        self.cn1 = Conv2dLayer(input_size[-1], 32, activation='relu')
        self.mp1 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.cn2 = Conv2dLayer(self.cn1.out_features, 64, activation='relu')
        self.mp2 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.cn3_1 = Conv2dLayer(self.cn2.out_features, 128, activation='relu')
        self.cn3_2 = Conv2dLayer(self.cn3_1.out_features, 128, activation='relu')
        self.mp3 = torch.nn.MaxPool2d(kernel_size=(2, 2))

        self.cn4_1 = Conv2dLayer(self.cn3_2.out_features, 128, activation='relu')
        self.cn4_2 = Conv2dLayer(self.cn4_1.out_features, 128, activation='relu')

        self.cn5_1 = Conv2dLayer(self.cn4_2.out_features, 256, activation='relu')
        self.cn5_2 = Conv2dLayer(self.cn5_1.out_features, 256, activation='relu')

        self.drop1 = torch.nn.Dropout(p=0.5)

        r = 2 ** 3
        fc1_input_size = self.cn5_2.out_features * input_size[0] // r * input_size[1] // r
        self.fc1 = FullyConnectedLayer(fc1_input_size, 256, activation='relu')

        self.drop2 = torch.nn.Dropout(p=0.5)
        self.fc2 = FullyConnectedLayer(self.fc1.out_features, 128, activation='relu')

        self.fc3 = FullyConnectedLayer(self.fc2.out_features, num_out_classes, activation=None)

    def forward(self, input_tensor):
        x = self.cn1(input_tensor)
        x = self.mp1(x)

        x = self.cn2(x)
        x = self.mp2(x)

        x = self.cn3_1(x)
        x = self.cn3_2(x)
        x = self.mp3(x)

        x = self.cn4_1(x)
        x = self.cn4_2(x)

        x = self.cn5_1(x)
        x = self.cn5_2(x)

        x = torch.flatten(x, start_dim=1)
        x = self.fc1(x)
        x = self.drop1(x)

        x = self.fc2(x)
        x = self.drop2(x)

        output = self.fc3(x)
        return output



import argparse

from pytorchDL.tasks.image_classification.evaluator import Evaluator


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--testDataDir', type=str, required=True, help='Directory containing the test data')
    parser.add_argument('--ckptPath', type=str, required=True, help='Path to checkpoint to evaluate')
    parser.add_argument('--device', type=str, required=False, default='cpu', choices=['cpu', 'gpu'],
                        help='Device in which to run evaluation')

    parser.add_argument('--outDir', type=str, required=True, help='Directory to store the test results')

    parser.add_argument('--batchSize', type=int, required=False, default=1, help='Batch size')
    parser.add_argument('-n', '--numProc', type=int, required=False, default=4,
                        help='Number of parallel processes for data loading')

    return parser.parse_args()


def main():
    args = parse_args()
    evaluator = Evaluator(test_data_dir=args.testDataDir,
                          out_dir=args.outDir,
                          ckpt_path=args.ckptPath,
                          batch_size=args.batchSize,
                          device=args.device,
                          num_proc=args.numProc)
    evaluator.run_testing()


if __name__ == '__main__':
    main()
